basisGradient[0] = -(1.0 - ss) / 4.0;
basisGradient[1] = -(1.0 - rr) / 4.0;
basisGradient[2] = 0.0;

basisGradient[3] = (1.0 - ss) / 4.0;
basisGradient[4] = -(1.0 + rr) / 4.0;
basisGradient[5] = 0.0;

basisGradient[6] = (1.0 + ss) / 4.0;
basisGradient[7] = (1.0 + rr) / 4.0;
basisGradient[8] = 0.0;

basisGradient[9] = -(1.0 + ss) / 4.0;
basisGradient[10] = (1.0 - rr) / 4.0;
basisGradient[11] = 0.0;
