#ifndef DIY_MPI_HPP
#define DIY_MPI_HPP

#include "mpi/config.hpp"
#include "mpi/datatypes.hpp"
#include "mpi/environment.hpp"
#include "mpi/optional.hpp"
#include "mpi/status.hpp"
#include "mpi/request.hpp"
#include "mpi/point-to-point.hpp"
#include "mpi/communicator.hpp"
#include "mpi/collectives.hpp"
#include "mpi/io.hpp"
#include "mpi/window.hpp"

#endif // DIY_MPI_HPP
