// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#ifndef TestLineWidget2EventLog_h
#define TestLineWidget2EventLog_h

// Divided into 3 parts because some compilers cannot support such large string
// literals
const char TestLineWidget2EventLog_p1[] = "# StreamVersion 1\n"
                                          "CharEvent 149 152 0 0 105 1 i\n"
                                          "MouseMoveEvent 149 152 0 0 0 0 i\n"
                                          "LeftButtonPressEvent 149 152 0 0 0 0 i\n"
                                          "MouseMoveEvent 146 152 0 0 0 0 i\n"
                                          "MouseMoveEvent 144 150 0 0 0 0 i\n"
                                          "MouseMoveEvent 40 85 0 0 0 0 i\n"
                                          "MouseMoveEvent 34 86 0 0 0 0 i\n"
                                          "MouseMoveEvent 29 88 0 0 0 0 i\n"
                                          "MouseMoveEvent 92 145 0 0 0 0 i\n"
                                          "MouseMoveEvent 94 145 0 0 0 0 i\n"
                                          "MouseMoveEvent 95 145 0 0 0 0 i\n"
                                          "MouseMoveEvent 96 145 0 0 0 0 i\n"
                                          "MouseMoveEvent 97 145 0 0 0 0 i\n"
                                          "MouseMoveEvent 99 145 0 0 0 0 i\n"
                                          "MouseMoveEvent 100 145 0 0 0 0 i\n"
                                          "MouseMoveEvent 99 146 0 0 0 0 i\n"
                                          "MouseMoveEvent 98 146 0 0 0 0 i\n"
                                          "MouseMoveEvent 96 146 0 0 0 0 i\n"
                                          "MouseMoveEvent 95 147 0 0 0 0 i\n"
                                          "MouseMoveEvent 94 147 0 0 0 0 i\n"
                                          "MouseMoveEvent 93 147 0 0 0 0 i\n"
                                          "MouseMoveEvent 91 148 0 0 0 0 i\n"
                                          "MouseMoveEvent 90 148 0 0 0 0 i\n"
                                          "MouseMoveEvent 89 148 0 0 0 0 i\n"
                                          "MouseMoveEvent 87 149 0 0 0 0 i\n"
                                          "MouseMoveEvent 86 150 0 0 0 0 i\n"
                                          "MouseMoveEvent 85 150 0 0 0 0 i\n"
                                          "MouseMoveEvent 83 151 0 0 0 0 i\n"
                                          "MouseMoveEvent 81 151 0 0 0 0 i\n"
                                          "MouseMoveEvent 80 151 0 0 0 0 i\n"
                                          "MouseMoveEvent 79 152 0 0 0 0 i\n"
                                          "MouseMoveEvent 78 152 0 0 0 0 i\n"
                                          "MouseMoveEvent 76 152 0 0 0 0 i\n"
                                          "MouseMoveEvent 75 152 0 0 0 0 i\n"
                                          "LeftButtonReleaseEvent 75 152 0 0 0 0 i\n"
                                          "MouseMoveEvent 76 150 0 0 0 0 i\n"
                                          "MouseMoveEvent 76 149 0 0 0 0 i\n"
                                          "MouseMoveEvent 76 148 0 0 0 0 i\n"
                                          "MouseMoveEvent 77 85 0 0 0 0 i\n"
                                          "MouseMoveEvent 77 84 0 0 0 0 i\n"
                                          "MouseMoveEvent 77 83 0 0 0 0 i\n"
                                          "MouseMoveEvent 77 81 0 0 0 0 i\n"
                                          "LeftButtonPressEvent 77 81 0 0 0 0 i\n"
                                          "MouseMoveEvent 81 83 0 0 0 0 i\n"
                                          "MouseMoveEvent 83 85 0 0 0 0 i\n"
                                          "MouseMoveEvent 84 86 0 0 0 0 i\n"
                                          "MouseMoveEvent 73 89 0 0 0 0 i\n"
                                          "MouseMoveEvent 75 89 0 0 0 0 i\n"
                                          "MouseMoveEvent 76 89 0 0 0 0 i\n"
                                          "MouseMoveEvent 77 89 0 0 0 0 i\n"
                                          "LeftButtonReleaseEvent 77 89 0 0 0 0 i\n"
                                          "MouseMoveEvent 78 91 0 0 0 0 i\n"
                                          "MouseMoveEvent 78 93 0 0 0 0 i\n"
                                          "MouseMoveEvent 78 94 0 0 0 0 i\n"
                                          "MouseMoveEvent 76 181 0 0 0 0 i\n"
                                          "MouseMoveEvent 76 182 0 0 0 0 i\n"
                                          "MouseMoveEvent 77 184 0 0 0 0 i\n"
                                          "MouseMoveEvent 83 214 0 0 0 0 i\n"
                                          "MouseMoveEvent 83 216 0 0 0 0 i\n"
                                          "LeftButtonPressEvent 83 216 0 0 0 0 i\n"
                                          "MouseMoveEvent 82 217 0 0 0 0 i\n"
                                          "MouseMoveEvent 81 217 0 0 0 0 i\n"
                                          "MouseMoveEvent 35 211 0 0 0 0 i\n"
                                          "MouseMoveEvent 33 214 0 0 0 0 i\n"
                                          "MouseMoveEvent 31 218 0 0 0 0 i\n"
                                          "MouseMoveEvent 29 220 0 0 0 0 i\n"
                                          "MouseMoveEvent 28 223 0 0 0 0 i\n"
                                          "MouseMoveEvent 28 224 0 0 0 0 i\n"
                                          "MouseMoveEvent 296 133 0 0 0 0 i\n"
                                          "MouseMoveEvent 290 129 0 0 0 0 i\n"
                                          "MouseMoveEvent 284 127 0 0 0 0 i\n"
                                          "MouseMoveEvent 277 125 0 0 0 0 i\n"
                                          "MouseMoveEvent 85 197 0 0 0 0 i\n"
                                          "MouseMoveEvent 84 198 0 0 0 0 i\n"
                                          "MouseMoveEvent 82 199 0 0 0 0 i\n"
                                          "MouseMoveEvent 81 199 0 0 0 0 i\n"
                                          "MouseMoveEvent 88 203 0 0 0 0 i\n"
                                          "MouseMoveEvent 87 204 0 0 0 0 i\n"
                                          "MouseMoveEvent 86 204 0 0 0 0 i\n"
                                          "MouseMoveEvent 84 204 0 0 0 0 i\n"
                                          "LeftButtonReleaseEvent 84 204 0 0 0 0 i\n"
                                          "MouseMoveEvent 83 204 0 0 0 0 i\n"
                                          "MouseMoveEvent 82 204 0 0 0 0 i\n"
                                          "MouseMoveEvent 81 204 0 0 0 0 i\n"
                                          "MouseMoveEvent 80 205 0 0 0 0 i\n"
                                          "MouseMoveEvent 79 206 0 0 0 0 i\n"
                                          "KeyPressEvent 79 206 0 0 120 1 x\n"
                                          "CharEvent 79 206 0 0 120 1 x\n"
                                          "LeftButtonPressEvent 79 206 0 0 0 0 x\n"
                                          "MouseMoveEvent 77 207 0 0 0 0 x\n"
                                          "MouseMoveEvent 74 207 0 0 0 0 x\n"
                                          "MouseMoveEvent 73 208 0 0 0 0 x\n"
                                          "MouseMoveEvent 72 208 0 0 0 0 x\n"
                                          "MouseMoveEvent 70 209 0 0 0 0 x\n"
                                          "MouseMoveEvent 68 209 0 0 0 0 x\n"
                                          "MouseMoveEvent 65 235 0 0 0 0 x\n"
                                          "MouseMoveEvent 68 235 0 0 0 0 x\n";

const char TestLineWidget2EventLog_p2[] = "MouseMoveEvent 72 235 0 0 0 0 x\n"
                                          "MouseMoveEvent 76 234 0 0 0 0 x\n"
                                          "MouseMoveEvent 251 267 0 0 0 0 x\n"
                                          "MouseMoveEvent 252 268 0 0 0 0 x\n"
                                          "MouseMoveEvent 254 269 0 0 0 0 x\n"
                                          "MouseMoveEvent 255 270 0 0 0 0 x\n"
                                          "MouseMoveEvent 256 270 0 0 0 0 x\n"
                                          "MouseMoveEvent 80 255 0 0 0 0 x\n"
                                          "MouseMoveEvent 82 255 0 0 0 0 x\n"
                                          "KeyReleaseEvent 82 255 0 0 120 1 x\n"
                                          "LeftButtonReleaseEvent 82 255 0 0 0 0 x\n"
                                          "MouseMoveEvent 83 254 0 0 0 0 x\n"
                                          "MouseMoveEvent 84 254 0 0 0 0 x\n"
                                          "MouseMoveEvent 80 205 0 0 0 0 x\n"
                                          "MouseMoveEvent 82 205 0 0 0 0 x\n"
                                          "MouseMoveEvent 83 205 0 0 0 0 x\n"
                                          "LeftButtonPressEvent 83 205 0 0 0 0 x\n"
                                          "KeyPressEvent 83 205 0 0 121 1 y\n"
                                          "CharEvent 83 205 0 0 121 1 y\n"
                                          "MouseMoveEvent 82 205 0 0 0 0 y\n"
                                          "MouseMoveEvent 81 206 0 0 0 0 y\n"
                                          "MouseMoveEvent 80 207 0 0 0 0 y\n"
                                          "MouseMoveEvent 95 293 0 0 0 0 y\n"
                                          "MouseMoveEvent 96 296 0 0 0 0 y\n"
                                          "MouseMoveEvent 126 233 0 0 0 0 y\n"
                                          "MouseMoveEvent 128 212 0 0 0 0 y\n"
                                          "MouseMoveEvent 129 210 0 0 0 0 y\n"
                                          "MouseMoveEvent 129 208 0 0 0 0 y\n"
                                          "MouseMoveEvent 129 207 0 0 0 0 y\n"
                                          "MouseMoveEvent 129 206 0 0 0 0 y\n"
                                          "KeyReleaseEvent 129 206 0 0 121 1 y\n"
                                          "LeftButtonReleaseEvent 129 206 0 0 0 0 y\n"
                                          "MouseMoveEvent 128 205 0 0 0 0 y\n"
                                          "MouseMoveEvent 126 205 0 0 0 0 y\n"
                                          "MouseMoveEvent 124 207 0 0 0 0 y\n"
                                          "MouseMoveEvent 86 210 0 0 0 0 y\n"
                                          "MouseMoveEvent 84 210 0 0 0 0 y\n"
                                          "MouseMoveEvent 83 210 0 0 0 0 y\n"
                                          "MouseMoveEvent 82 210 0 0 0 0 y\n"
                                          "LeftButtonPressEvent 82 210 0 0 0 0 y\n"
                                          "KeyPressEvent 82 210 0 0 122 1 z\n"
                                          "CharEvent 82 210 0 0 122 1 z\n"
                                          "MouseMoveEvent 81 210 0 0 0 0 z\n"
                                          "MouseMoveEvent 78 210 0 0 0 0 z\n"
                                          "MouseMoveEvent 77 210 0 0 0 0 z\n"
                                          "MouseMoveEvent 74 210 0 0 0 0 z\n"
                                          "MouseMoveEvent 86 174 0 0 0 0 z\n"
                                          "MouseMoveEvent 78 198 0 0 0 0 z\n"
                                          "MouseMoveEvent 78 199 0 0 0 0 z\n"
                                          "MouseMoveEvent 78 201 0 0 0 0 z\n"
                                          "KeyReleaseEvent 78 201 0 0 122 1 z\n"
                                          "LeftButtonReleaseEvent 78 201 0 0 0 0 z\n"
                                          "MouseMoveEvent 80 196 0 0 0 0 z\n"
                                          "MouseMoveEvent 80 193 0 0 0 0 z\n"
                                          "MouseMoveEvent 81 190 0 0 0 0 z\n"
                                          "MouseMoveEvent 83 184 0 0 0 0 z\n"
                                          "MouseMoveEvent 85 82 0 0 0 0 z\n"
                                          "MouseMoveEvent 85 81 0 0 0 0 z\n"
                                          "MouseMoveEvent 84 78 0 0 0 0 z\n"
                                          "MouseMoveEvent 71 84 0 0 0 0 z\n"
                                          "MouseMoveEvent 71 85 0 0 0 0 z\n"
                                          "MouseMoveEvent 72 86 0 0 0 0 z\n"
                                          "MouseMoveEvent 73 86 0 0 0 0 z\n"
                                          "LeftButtonPressEvent 73 86 0 0 0 0 z\n"
                                          "KeyPressEvent 73 86 0 0 120 1 x\n"
                                          "CharEvent 73 86 0 0 120 1 x\n"
                                          "MouseMoveEvent 72 86 0 0 0 0 x\n"
                                          "MouseMoveEvent 70 86 0 0 0 0 x\n"
                                          "MouseMoveEvent 66 85 0 0 0 0 x\n"
                                          "MouseMoveEvent 60 84 0 0 0 0 x\n"
                                          "MouseMoveEvent 59 83 0 0 0 0 x\n"
                                          "MouseMoveEvent 55 83 0 0 0 0 x\n"
                                          "MouseMoveEvent 271 29 0 0 0 0 x\n"
                                          "MouseMoveEvent 272 29 0 0 0 0 x\n"
                                          "MouseMoveEvent 158 17 0 0 0 0 x\n"
                                          "MouseMoveEvent 156 17 0 0 0 0 x\n"
                                          "MouseMoveEvent 152 17 0 0 0 0 x\n"
                                          "MouseMoveEvent 78 17 0 0 0 0 x\n"
                                          "MouseMoveEvent 77 17 0 0 0 0 x\n"
                                          "MouseMoveEvent 76 17 0 0 0 0 x\n"
                                          "MouseMoveEvent 75 17 0 0 0 0 x\n"
                                          "MouseMoveEvent 73 17 0 0 0 0 x\n"
                                          "KeyReleaseEvent 73 17 0 0 120 1 x\n"
                                          "LeftButtonReleaseEvent 73 17 0 0 0 0 x\n"
                                          "MouseMoveEvent 72 18 0 0 0 0 x\n"
                                          "MouseMoveEvent 72 19 0 0 0 0 x\n"
                                          "MouseMoveEvent 72 22 0 0 0 0 x\n"
                                          "MouseMoveEvent 72 23 0 0 0 0 x\n"
                                          "MouseMoveEvent 78 80 0 0 0 0 x\n"
                                          "MouseMoveEvent 78 81 0 0 0 0 x\n"
                                          "MouseMoveEvent 78 82 0 0 0 0 x\n"
                                          "MouseMoveEvent 78 83 0 0 0 0 x\n"
                                          "MouseMoveEvent 78 84 0 0 0 0 x\n"
                                          "MouseMoveEvent 79 85 0 0 0 0 x\n"
                                          "LeftButtonPressEvent 79 85 0 0 0 0 x\n"
                                          "KeyPressEvent 79 85 0 0 121 1 y\n"
                                          "CharEvent 79 85 0 0 121 1 y\n"
                                          "MouseMoveEvent 79 88 0 0 0 0 y\n";

const char TestLineWidget2EventLog_p3[] = "MouseMoveEvent 79 94 0 0 0 0 y\n"
                                          "MouseMoveEvent 76 102 0 0 0 0 y\n"
                                          "MouseMoveEvent 74 110 0 0 0 0 y\n"
                                          "MouseMoveEvent 73 117 0 0 0 0 y\n"
                                          "MouseMoveEvent 71 124 0 0 0 0 y\n"
                                          "MouseMoveEvent 68 131 0 0 0 0 y\n"
                                          "MouseMoveEvent 67 136 0 0 0 0 y\n"
                                          "MouseMoveEvent 39 63 0 0 0 0 y\n"
                                          "MouseMoveEvent 39 64 0 0 0 0 y\n"
                                          "MouseMoveEvent 39 65 0 0 0 0 y\n"
                                          "MouseMoveEvent 39 66 0 0 0 0 y\n"
                                          "MouseMoveEvent 40 67 0 0 0 0 y\n"
                                          "MouseMoveEvent 40 69 0 0 0 0 y\n"
                                          "LeftButtonReleaseEvent 40 69 0 0 0 0 y\n"
                                          "MouseMoveEvent 42 70 0 0 0 0 y\n"
                                          "MouseMoveEvent 44 70 0 0 0 0 y\n"
                                          "MouseMoveEvent 48 70 0 0 0 0 y\n"
                                          "MouseMoveEvent 56 70 0 0 0 0 y\n"
                                          "MouseMoveEvent 59 70 0 0 0 0 y\n"
                                          "MouseMoveEvent 62 70 0 0 0 0 y\n"
                                          "MouseMoveEvent 63 70 0 0 0 0 y\n"
                                          "KeyReleaseEvent 63 70 0 0 121 1 y\n"
                                          "MouseMoveEvent 64 69 0 0 0 0 y\n"
                                          "MouseMoveEvent 65 69 0 0 0 0 y\n"
                                          "MouseMoveEvent 67 69 0 0 0 0 y\n"
                                          "MouseMoveEvent 68 69 0 0 0 0 y\n"
                                          "MouseMoveEvent 69 68 0 0 0 0 y\n"
                                          "MouseMoveEvent 70 68 0 0 0 0 y\n"
                                          "MouseMoveEvent 72 67 0 0 0 0 y\n"
                                          "MouseMoveEvent 73 67 0 0 0 0 y\n"
                                          "MouseMoveEvent 74 67 0 0 0 0 y\n"
                                          "MouseMoveEvent 75 67 0 0 0 0 y\n"
                                          "MouseMoveEvent 77 67 0 0 0 0 y\n"
                                          "MouseMoveEvent 78 67 0 0 0 0 y\n"
                                          "MouseMoveEvent 79 67 0 0 0 0 y\n"
                                          "LeftButtonPressEvent 79 67 0 0 0 0 y\n"
                                          "KeyPressEvent 79 67 0 0 122 1 z\n"
                                          "CharEvent 79 67 0 0 122 1 z\n"
                                          "MouseMoveEvent 78 67 0 0 0 0 z\n"
                                          "MouseMoveEvent 76 67 0 0 0 0 z\n"
                                          "MouseMoveEvent 71 67 0 0 0 0 z\n"
                                          "MouseMoveEvent 64 67 0 0 0 0 z\n"
                                          "MouseMoveEvent 57 67 0 0 0 0 z\n"
                                          "MouseMoveEvent 64 50 0 0 0 0 z\n"
                                          "MouseMoveEvent 67 50 0 0 0 0 z\n"
                                          "MouseMoveEvent 226 68 0 0 0 0 z\n"
                                          "MouseMoveEvent 225 68 0 0 0 0 z\n"
                                          "MouseMoveEvent 121 66 0 0 0 0 z\n"
                                          "MouseMoveEvent 119 67 0 0 0 0 z\n"
                                          "MouseMoveEvent 117 67 0 0 0 0 z\n"
                                          "MouseMoveEvent 92 73 0 0 0 0 z\n"
                                          "MouseMoveEvent 90 74 0 0 0 0 z\n"
                                          "MouseMoveEvent 88 74 0 0 0 0 z\n"
                                          "MouseMoveEvent 87 74 0 0 0 0 z\n"
                                          "MouseMoveEvent 86 74 0 0 0 0 z\n"
                                          "MouseMoveEvent 84 74 0 0 0 0 z\n"
                                          "LeftButtonReleaseEvent 84 74 0 0 0 0 z\n"
                                          "KeyReleaseEvent 84 74 0 0 122 1 z\n"
                                          "MouseMoveEvent 84 77 0 0 0 0 z\n"
                                          "MouseMoveEvent 85 81 0 0 0 0 z\n"
                                          "MouseMoveEvent 85 85 0 0 0 0 z\n"
                                          "MouseMoveEvent 86 86 0 0 0 0 z\n"
                                          "MouseMoveEvent 86 90 0 0 0 0 z\n"
                                          "MouseMoveEvent 86 93 0 0 0 0 z\n"
                                          "MouseMoveEvent 81 134 0 0 0 0 z\n"
                                          "MouseMoveEvent 80 136 0 0 0 0 z\n"
                                          "MouseMoveEvent 80 138 0 0 0 0 z\n"
                                          "MouseMoveEvent 80 139 0 0 0 0 z\n"
                                          "MouseMoveEvent 80 140 0 0 0 0 z\n"
                                          "MouseMoveEvent 80 141 0 0 0 0 z\n"
                                          "LeftButtonPressEvent 80 141 0 0 0 0 z\n"
                                          "KeyPressEvent 80 141 0 0 120 1 x\n"
                                          "CharEvent 80 141 0 0 120 1 x\n"
                                          "MouseMoveEvent 80 140 0 0 0 0 x\n"
                                          "MouseMoveEvent 80 137 0 0 0 0 x\n"
                                          "MouseMoveEvent 80 134 0 0 0 0 x\n"
                                          "MouseMoveEvent 118 106 0 0 0 0 x\n"
                                          "MouseMoveEvent 120 106 0 0 0 0 x\n"
                                          "MouseMoveEvent 123 104 0 0 0 0 x\n"
                                          "MouseMoveEvent 126 104 0 0 0 0 x\n"
                                          "MouseMoveEvent 129 103 0 0 0 0 x\n"
                                          "MouseMoveEvent 227 74 0 0 0 0 x\n"
                                          "MouseMoveEvent 230 74 0 0 0 0 x\n"
                                          "MouseMoveEvent 232 73 0 0 0 0 x\n"
                                          "MouseMoveEvent 39 94 0 0 0 0 x\n"
                                          "MouseMoveEvent 78 118 0 0 0 0 x\n"
                                          "MouseMoveEvent 80 119 0 0 0 0 x\n"
                                          "MouseMoveEvent 81 120 0 0 0 0 x\n"
                                          "MouseMoveEvent 82 120 0 0 0 0 x\n"
                                          "MouseMoveEvent 84 121 0 0 0 0 x\n"
                                          "MouseMoveEvent 87 122 0 0 0 0 x\n"
                                          "MouseMoveEvent 88 122 0 0 0 0 x\n"
                                          "MouseMoveEvent 89 122 0 0 0 0 x\n"
                                          "KeyReleaseEvent 89 122 0 0 120 1 x\n"
                                          "LeftButtonReleaseEvent 89 122 0 0 0 0 x\n"
                                          "MouseMoveEvent 90 123 0 0 0 0 x\n"
                                          "MouseMoveEvent 90 124 0 0 0 0 x\n"
                                          "MouseMoveEvent 90 125 0 0 0 0 x\n"
                                          "MouseMoveEvent 91 127 0 0 0 0 x\n"
                                          "MouseMoveEvent 92 128 0 0 0 0 x\n"
                                          "MouseMoveEvent 92 129 0 0 0 0 x\n"
                                          "MouseMoveEvent 92 130 0 0 0 0 x\n"
                                          "LeftButtonPressEvent 92 130 0 0 0 0 x\n"
                                          "KeyPressEvent 92 130 0 0 121 1 y\n"
                                          "CharEvent 92 130 0 0 121 1 y\n"
                                          "MouseMoveEvent 92 132 0 0 0 0 y\n"
                                          "MouseMoveEvent 92 134 0 0 0 0 y\n"
                                          "MouseMoveEvent 92 136 0 0 0 0 y\n"
                                          "MouseMoveEvent 92 139 0 0 0 0 y\n"
                                          "MouseMoveEvent 95 135 0 0 0 0 y\n"
                                          "MouseMoveEvent 95 127 0 0 0 0 y\n"
                                          "MouseMoveEvent 95 119 0 0 0 0 y\n"
                                          "MouseMoveEvent 105 66 0 0 0 0 y\n"
                                          "MouseMoveEvent 105 69 0 0 0 0 y\n"
                                          "MouseMoveEvent 105 71 0 0 0 0 y\n"
                                          "MouseMoveEvent 105 74 0 0 0 0 y\n"
                                          "MouseMoveEvent 107 126 0 0 0 0 y\n"
                                          "MouseMoveEvent 107 127 0 0 0 0 y\n"
                                          "MouseMoveEvent 107 129 0 0 0 0 y\n"
                                          "MouseMoveEvent 107 130 0 0 0 0 y\n"
                                          "MouseMoveEvent 107 131 0 0 0 0 y\n"
                                          "LeftButtonReleaseEvent 107 131 0 0 0 0 y\n"
                                          "KeyReleaseEvent 107 131 0 0 121 1 y\n"
                                          "MouseMoveEvent 106 132 0 0 0 0 y\n"
                                          "MouseMoveEvent 102 133 0 0 0 0 y\n"
                                          "MouseMoveEvent 96 136 0 0 0 0 y\n"
                                          "MouseMoveEvent 95 137 0 0 0 0 y\n"
                                          "MouseMoveEvent 92 138 0 0 0 0 y\n"
                                          "MouseMoveEvent 91 139 0 0 0 0 y\n"
                                          "MouseMoveEvent 91 140 0 0 0 0 y\n"
                                          "MouseMoveEvent 90 142 0 0 0 0 y\n"
                                          "MouseMoveEvent 88 142 0 0 0 0 y\n"
                                          "LeftButtonPressEvent 88 142 0 0 0 0 y\n"
                                          "KeyPressEvent 88 142 0 0 122 1 z\n"
                                          "CharEvent 88 142 0 0 122 1 z\n"
                                          "MouseMoveEvent 87 142 0 0 0 0 z\n"
                                          "MouseMoveEvent 83 140 0 0 0 0 z\n"
                                          "MouseMoveEvent 77 138 0 0 0 0 z\n"
                                          "MouseMoveEvent 72 137 0 0 0 0 z\n"
                                          "MouseMoveEvent 66 137 0 0 0 0 z\n"
                                          "MouseMoveEvent 60 137 0 0 0 0 z\n"
                                          "MouseMoveEvent 62 118 0 0 0 0 z\n"
                                          "MouseMoveEvent 66 118 0 0 0 0 z\n"
                                          "MouseMoveEvent 70 116 0 0 0 0 z\n"
                                          "MouseMoveEvent 74 115 0 0 0 0 z\n"
                                          "MouseMoveEvent 160 109 0 0 0 0 z\n"
                                          "MouseMoveEvent 162 109 0 0 0 0 z\n"
                                          "MouseMoveEvent 163 109 0 0 0 0 z\n"
                                          "MouseMoveEvent 61 146 0 0 0 0 z\n"
                                          "MouseMoveEvent 60 146 0 0 0 0 z\n"
                                          "MouseMoveEvent 59 146 0 0 0 0 z\n"
                                          "MouseMoveEvent 57 147 0 0 0 0 z\n"
                                          "MouseMoveEvent 56 147 0 0 0 0 z\n"
                                          "LeftButtonReleaseEvent 56 147 0 0 0 0 z\n"
                                          "KeyReleaseEvent 56 147 0 0 122 1 z\n"
                                          "MouseMoveEvent 57 148 0 0 0 0 z\n"
                                          "MouseMoveEvent 60 148 0 0 0 0 z\n"
                                          "MouseMoveEvent 63 148 0 0 0 0 z\n"
                                          "MouseMoveEvent 92 131 0 0 0 0 z\n"
                                          "MouseMoveEvent 92 130 0 0 0 0 z\n"
                                          "MouseMoveEvent 92 129 0 0 0 0 z\n"
                                          "LeftButtonPressEvent 92 129 0 0 0 0 z\n"
                                          "MouseMoveEvent 93 129 0 0 0 0 z\n"
                                          "MouseMoveEvent 94 130 0 0 0 0 z\n"
                                          "MouseMoveEvent 95 131 0 0 0 0 z\n"
                                          "MouseMoveEvent 96 132 0 0 0 0 z\n"
                                          "MouseMoveEvent 116 143 0 0 0 0 z\n"
                                          "MouseMoveEvent 116 141 0 0 0 0 z\n"
                                          "MouseMoveEvent 116 140 0 0 0 0 z\n"
                                          "MouseMoveEvent 116 139 0 0 0 0 z\n"
                                          "MouseMoveEvent 116 138 0 0 0 0 z\n"
                                          "MouseMoveEvent 115 137 0 0 0 0 z\n"
                                          "MouseMoveEvent 115 136 0 0 0 0 z\n"
                                          "MouseMoveEvent 115 135 0 0 0 0 z\n"
                                          "MouseMoveEvent 138 167 0 0 0 0 i\n"
                                          "MouseMoveEvent 139 168 0 0 0 0 i\n"
                                          "MouseMoveEvent 140 168 0 0 0 0 i\n"
                                          "MouseMoveEvent 141 168 0 0 0 0 i\n"
                                          "MouseMoveEvent 142 168 0 0 0 0 i\n"
                                          "MouseMoveEvent 156 172 0 0 0 0 i\n"
                                          "MouseMoveEvent 157 172 0 0 0 0 i\n"
                                          "MouseMoveEvent 158 171 0 0 0 0 i\n"
                                          "MouseMoveEvent 160 171 0 0 0 0 i\n"
                                          "MouseMoveEvent 162 170 0 0 0 0 i\n"
                                          "MouseMoveEvent 163 170 0 0 0 0 i\n"
                                          "MouseMoveEvent 164 170 0 0 0 0 i\n"
                                          "MouseMoveEvent 165 170 0 0 0 0 i\n"
                                          "LeftButtonReleaseEvent 115 135 0 0 0 0 z\n"
                                          "MouseMoveEvent 115 134 0 0 0 0 z\n";

#endif // TestLineWidget2EventLog_h
