// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause

#ifndef vtkLabeledDatatypeDefinitions_h
#define vtkLabeledDatatypeDefinitions_h

#include "vtkABINamespace.h" // For VTK_ABI_NAMESPACE_*

VTK_ABI_NAMESPACE_BEGIN
#define VTK_LABEL_IDS 0
#define VTK_LABEL_SCALARS 1
#define VTK_LABEL_VECTORS 2
#define VTK_LABEL_NORMALS 3
#define VTK_LABEL_TCOORDS 4
#define VTK_LABEL_TENSORS 5
#define VTK_LABEL_FIELD_DATA 6
VTK_ABI_NAMESPACE_END

#endif
