// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#ifndef uses_vtk_h
#define uses_vtk_h

#include "vtk_using_lib_export.h"

#include <string>

std::string VTK_USING_LIB_EXPORT vtkObject_class_name();

#endif
