// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause

namespace detail
{
int performTest(int argc, char* argv[], const char* filename);
}
