// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#include "vtkAbstractPointLocator.h"
#include "vtkVoronoiKernel.h"

//------------------------------------------------------------------------------
VTK_ABI_NAMESPACE_BEGIN
vtkVoronoiKernel::vtkVoronoiKernel() {}

//------------------------------------------------------------------------------
vtkVoronoiKernel::~vtkVoronoiKernel() {}

//------------------------------------------------------------------------------
void vtkVoronoiKernel::vtkIdType ComputeWeights(
  double x[3], vtkIdList* pIds, vtkDoubleArray* weights)
{
}

//------------------------------------------------------------------------------
void vtkVoronoiKernel::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os, indent);
}
VTK_ABI_NAMESPACE_END
