// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#ifndef vtkTcl_h
#define vtkTcl_h

#include "vtkTkAppInitConfigure.h"

#ifdef VTK_TCL_TK_STATIC
#ifndef STATIC_BUILD
#define STATIC_BUILD
#endif
#endif

#include <tcl.h>

#endif
// VTK-HeaderTest-Exclude: vtkTcl.h
