// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#import <Cocoa/Cocoa.h>

@interface MyWindowController : NSWindowController

// Designated initializer
- (instancetype)init /*NS_DESIGNATED_INITIALIZER*/;

@end
