.canvas {
    left: 0;
    top: 0;
    position: absolute;
    display: none;
}

.container {
    display: flex;
    flex-direction: column;
    width: 100%;
    height: 100vh;
}

.center {
    margin: 0;
    position: absolute;
    top: 50%;
    left: 50%;
    -ms-transform: translate(-50%, -50%);
    transform: translate(-50%, -50%);
}
