// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-FileCopyrightText: Copyright (c) Kitware, Inc.
// SPDX-FileCopyrightText: Copyright 2012 Sandia Corporation.
// SPDX-License-Identifier: LicenseRef-BSD-3-Clause-Sandia-USGov

#ifndef vtkmlib_vtkmInitializer_h
#define vtkmlib_vtkmInitializer_h

#include "vtkAcceleratorsVTKmCoreModule.h"

VTK_ABI_NAMESPACE_BEGIN
struct VTKACCELERATORSVTKMCORE_EXPORT vtkmInitializer
{
  vtkmInitializer();
};

VTK_ABI_NAMESPACE_END
#endif
